!***************************************************
! Two dimensional Split of Jacobi Iteration
! Qun Liu  contact by liu-q14@mails.tsinghua.edu.cn
! Time: 2015-3-28
!***************************************************

program main

    use mpi
    implicit none

    integer, parameter:: xtotalsize = 16                ! 总网格数目为xtotalsize*ytotalsize
    integer, parameter:: ytotalsize = 16              
    integer, parameter:: xproc      = 4                 ! x 方向的进程数
    integer, parameter:: yproc      = 4                 ! y 方向的进程数
    integer, parameter:: steps      = 100               ! 迭代步数
    integer, parameter:: xsize      = xtotalsize/xproc  ! x 方向的格点数
    integer, parameter:: ysize      = ytotalsize/yproc  ! y 方向的格点数
    real a(ysize+2, xsize+2), b(ysize+2, xsize+2)        
    integer n, i, j, myid, numprocs, ierr, left, right, up, down
    integer tag1, tag2, tag3, tag4, ROWTYPE
    integer begin_col, end_col, begin_row, end_row
    integer status(MPI_STATUS_SIZE)

    ! MPI Initialization
    call MPI_INIT(ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD, myid, ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, numprocs, ierr)
    
    if(numprocs /= xproc * yproc .or. mod(xtotalsize, xproc) /= 0 .or. mod(ytotalsize, yproc) /= 0) then
        write(*,*) "Error: mpirun -n p xxx should be: mpi -n ", xproc*yproc, "xxx or grid scale error."
        call MPI_FINALIZE(ierr)
        stop
    end if
    
    ! Two Dimensional Arrray Iniatilization 
    do j = 1,xsize+2                
        do i = 1,ysize+2
            a(i, j) = 0.0
        end do
    end do
    
    if (myid < xproc) then                  ! Up Boundary Condition
        do j = 2,xsize+1
            a(2, j) = 8.0
        end do
    end if

    if (myid >= xproc*(yproc-1)) then       ! Down Boundary Condition
        do j = 2,xsize+1
            a(ysize+1, j) = 8.0
        end do
    end if

    if (mod(myid, xproc) == 0) then         ! Left Boundary Condition
        do i = 2,ysize+1
            a(i, 2) = 8.0
        end do
    end if

    if (mod(myid, xproc) == xproc-1) then   ! Right Boundary Condition
        do i = 2,ysize+1
            a(i, xsize+1) = 8.0
        end do
    end if
    
    ! Set left, right, up and down process id
    if (mod(myid, xproc) == 0) then
        left = MPI_PROC_NULL
    else
        left = myid-1
    end if

    if (mod(myid, xproc) == xproc-1) then
        right = MPI_PROC_NULL
    else
        right = myid+1
    end if 

    if (myid < xproc) then
        up = MPI_PROC_NULL
    else
        up = myid-xproc
    end if 

    if (myid >= xproc*(yproc-1)) then
        down = MPI_PROC_NULL
    else
        down = myid+xproc
    end if 
    
    tag1 = 1
    tag2 = 2
    tag3 = 3
    tag4 = 4

    ! Define new data type
    call MPI_TYPE_VECTOR(xsize, 1, ysize+2, MPI_REAL, ROWTYPE, ierr)
    call MPI_TYPE_COMMIT(ROWTYPE, ierr)

    do n = 1, steps
        ! data from left to right
        call MPI_SENDRECV(a(2,xsize+1), ysize, MPI_REAL, right, tag1, &
            a(2,1), ysize, MPI_REAL, left, tag1, MPI_COMM_WORLD, status, ierr)

        ! data from right to left
        call MPI_SENDRECV(a(2,2), ysize, MPI_REAL, left, tag2, &
            a(2,xsize+2), ysize, MPI_REAL, right, tag2, MPI_COMM_WORLD, status, ierr)

        ! data from down to up 
        call MPI_SENDRECV(a(2,2), 1, ROWTYPE, up, tag3, &
            a(ysize+2, 2), 1, ROWTYPE, down, tag3, MPI_COMM_WORLD, status, ierr)

        ! data from up to down 
        call MPI_SENDRECV(a(ysize+1,2), 1, ROWTYPE, down, tag4, &
            a(1,2), 1, ROWTYPE, up, tag4, MPI_COMM_WORLD, status, ierr)

        ! Determine the begin and end col and row
        begin_col = 2
        end_col = xsize+1
        begin_row = 2
        end_row = ysize+1
        
        if (mod(myid, xproc) == 0) then
            begin_col = 3
        end if

        if (mod(myid, xproc) == xproc-1) then
            end_col = xsize
        end if

        if (myid < xproc) then
            begin_row = 3
        end if
        
        if (myid >= xproc*(yproc-1)) then
            end_row = ysize
        end if

        do j = begin_col, end_col
            do i = begin_row, end_row 
                b(i, j) = (a(i,j+1)+a(i,j-1)+a(i+1,j)+a(i-1,j))*0.25
            end do
        end do

        do j = begin_col, end_col
            do i = begin_row, end_row 
                a(i,j) = b(i, j)
            end do
        end do

    end do

    write(*,'(I2, 4(4f6.2,/))') myid, ((a(i,j), j=2,xsize+1),i=2,ysize+1)
    !write(*,'(I2,/, 16(4f6.2,/))') myid, ((a(i,j), j=2,xsize+1),i=2,ysize+1)
    !write(*,'(I2,/, 10(80f6.2,/))') myid, ((a(i,j), j=2,xsize+1),i=2,ysize+1)

    call MPI_FINALIZE(ierr)

end program main 